/*
 * Created on Nov 5, 2011
 *
 */
package org.gk.pathwaylayout;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;

import org.gk.graphEditor.PathwayEditor;
import org.gk.model.GKInstance;
import org.gk.persistence.DiagramGKBReader;
import org.gk.persistence.GKBWriter;
import org.gk.persistence.MySQLAdaptor;
import org.gk.persistence.Project;
import org.gk.render.RenderablePathway;
import org.junit.Test;

import com.sun.xml.internal.messaging.saaj.util.ByteOutputStream;

/**
 * This class is used to generate an XML for a PathwayDiagram instance based on stored
 * XML string in the database. The generated XML should be sufficient to be used for display
 * a pahway diagram as in the curator tool. In other words, all rendering information should
 * be included in the generated XML.
 * @author gwu
 *
 */
public class PathwayDiagramXMLGenerator {
    
    public PathwayDiagramXMLGenerator() {
    }
    
    /**
     * Generate an XML file for a specified PathwayDiagram instance.
     * @param pd
     * @param file
     * @throws Exception
     */
    public void generateXMLForPathwayDiagram(GKInstance pd, 
                                             File file) throws Exception {
        generateXMLForPathwayDiagram(pd, new FileOutputStream(file));
    }
    
    /**
     * Generate an XML for a specified PathwayDiagram to an OutputStream object.
     * @param pd
     * @param os
     * @throws Exception
     */
    public void generateXMLForPathwayDiagram(GKInstance pd, 
                                             OutputStream os) throws Exception {
        DiagramGKBReader reader = new DiagramGKBReader();
        RenderablePathway pathway = reader.openDiagram(pd);
        // Force to make sure all points are correct validated
        PathwayEditor pathwayEditor = new PathwayEditor();
        pathwayEditor.setRenderable(pathway);
        new PathwayDiagramGeneratorViaAT().paintOnImage(pathwayEditor);
        GKBWriter writer = new GKBWriter();
        writer.setNeedRegistryCheck(false);
        writer.save(new Project(pathway), 
                    os);
    }
    
    /**
     * Generate an XML String object for a specified PathwayDiagram instance.
     * @param pd
     * @return
     * @throws Exception
     */
    public String generateXMLForPathwayDigram(GKInstance pd) throws Exception {
        ByteOutputStream bos = new ByteOutputStream();
        generateXMLForPathwayDiagram(pd, bos);
        String text = bos.toString();
        return text;
    }
    
    @Test
    public void testGenerateXMLForPathwayDigram() throws Exception {
        MySQLAdaptor dba = new MySQLAdaptor("localhost",
                                            "gk_current_ver38",
                                            "root",
                                            "macmysql01");
        // EGFR Pathway diagram
        Long pdDbId = 507988L;
        GKInstance pdInst = dba.fetchInstance(pdDbId);
        String xmlText = generateXMLForPathwayDigram(pdInst);
        System.out.println(xmlText);
    }
    
}
