/*
 * NodeConnectInfo.java
 *
 * Created on June 20, 2003, 4:17 PM
 */

package org.gk.render;

import java.awt.Point;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
/**
 * This class contains information related to nodes connecting to reactions.
 * @author  wgm
 */
public abstract class ConnectInfo implements Serializable {
    // A list of ConnectWidgets
    protected java.util.List connectWidgets;
    
    /** Creates a new instance of NodeConnectInfo */
    public ConnectInfo() {
    }
    
    public void setConnectWidgets(java.util.List widgets) {
        this.connectWidgets = widgets;
    }
    
    public java.util.List getConnectWidgets() {
        return this.connectWidgets;
    }
    
    public void addConnectWidget(ConnectWidget widget) {
        if (connectWidgets == null)
            connectWidgets = new ArrayList();
        if (!connectWidgets.contains(widget))
            connectWidgets.add(widget);
    }
    
    public void removeConnectWidget(ConnectWidget widget) {
        if (connectWidgets != null)
            connectWidgets.remove(widget);
    }
    
    public ConnectWidget searchConnectWidget(Point p) {
        if (connectWidgets == null)
            return null;
        ConnectWidget widget = null;
        for (Iterator it = connectWidgets.iterator(); it.hasNext();) {
            widget = (ConnectWidget) it.next();
            if (widget.getPoint() == p)
                return widget;
        }
        return null;
    }
    
    public void invalidate() {
        if (connectWidgets != null) {
            ConnectWidget widget = null;
            HyperEdge edge = null;
            for (Iterator it = connectWidgets.iterator(); it.hasNext();) {
                widget = (ConnectWidget)it.next();
                if (widget.isInvalidate())
                	continue;
                widget.invalidate();
				edge = widget.getEdge();
				// Have to invalidate both ends
				if (edge.getBackbonePoints().size() == 2) {
					edge.getConnectInfo().invalidate();
				}
            }
        }
    }
    
    public void validate() {
        if (connectWidgets == null)
            return;
        ConnectWidget widget = null;
        for (Iterator it = connectWidgets.iterator(); it.hasNext();) {
            widget = (ConnectWidget)it.next();
            widget.validate();
        }
    }    
    
    public abstract void clear();
}
